/* Copyright (C) 2009 Sun Microsystems

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef STDC_HEADERS
# include <stdio.h>
#endif
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif
#include <stdio.h>

#include <libdrizzleclient/drizzle.h>
#include <drizzled/error.h>

#include "sb_options.h"
#include "db_driver.h"


/* Check if we should use the TYPE= (for 3.23) or ENGINE= syntax */
#define ENGINE_CLAUSE "ENGINE"

#define DEBUG(format, ...) do { if (db_globals.debug) log_text(LOG_DEBUG, format, __VA_ARGS__); } while (0)

/* Drizzle driver arguments */

static sb_arg_t drizzleclient_drv_args[] =
{
  {"drizzle-host", "Drizzle server host", SB_ARG_TYPE_LIST, "localhost"},
  {"drizzle-port", "Drizzle server port", SB_ARG_TYPE_INT, "3306"},
  {"drizzle-socket", "Drizzle socket", SB_ARG_TYPE_STRING, NULL},
  {"drizzle-user", "Drizzle user", SB_ARG_TYPE_STRING, "sbtest"},
  {"drizzle-password", "Drizzle password", SB_ARG_TYPE_STRING, ""},
  {"drizzle-db", "Drizzle database name", SB_ARG_TYPE_STRING, "sbtest"},
  {"drizzle-table-engine", "storage engine to use for the test table {myisam,innodb,bdb,heap,ndbcluster,federated}",
   SB_ARG_TYPE_STRING, "innodb"},
  {"drizzle-engine-trx", "whether storage engine used is transactional or not {yes,no,auto}",
   SB_ARG_TYPE_STRING, "auto"},
  {"myisam-max-rows", "max-rows parameter for MyISAM tables", SB_ARG_TYPE_INT, "1000000"},
  {"drizzle-create-options", "additional options passed to CREATE TABLE", SB_ARG_TYPE_STRING, ""},
  
  {NULL, NULL, SB_ARG_TYPE_NULL, NULL}
};

typedef enum
{
  ENGINE_TRX_YES,
  ENGINE_TRX_NO,
  ENGINE_TRX_AUTO
} drizzleclient_drv_trx_t;

typedef struct
{
  sb_list_t          *hosts;
  unsigned int       port;
  char               *socket;
  char               *user;
  char               *password;
  char               *db;
  unsigned int       myisam_max_rows;
  drizzleclient_drv_trx_t    engine_trx;
  char               *create_options;
} drizzleclient_drv_args_t;

#ifdef HAVE_PS
/* Structure used for DB-to-Drizzle bind types map */

typedef struct
{
  db_bind_type_t   db_type;
  int              my_type;
} db_drizzleclient_bind_map_t;

/* DB-to-Drizzle bind types map */
db_drizzleclient_bind_map_t db_drizzleclient_bind_map[] =
{
  {DB_TYPE_TINYINT,   DRIZZLE_TYPE_TINY},
  {DB_TYPE_SMALLINT,  DRIZZLE_TYPE_SHORT},
  {DB_TYPE_INT,       DRIZZLE_TYPE_LONG},
  {DB_TYPE_BIGINT,    DRIZZLE_TYPE_LONGLONG},
  {DB_TYPE_FLOAT,     DRIZZLE_TYPE_FLOAT},
  {DB_TYPE_DOUBLE,    DRIZZLE_TYPE_DOUBLE},
  {DB_TYPE_DATETIME,  DRIZZLE_TYPE_DATETIME},
  {DB_TYPE_TIMESTAMP, DRIZZLE_TYPE_TIMESTAMP},
  {DB_TYPE_CHAR,      DRIZZLE_TYPE_STRING},
  {DB_TYPE_VARCHAR,   DRIZZLE_TYPE_VAR_STRING},
  {DB_TYPE_NONE,      0}
};

#endif /* HAVE_PS */

/* Drizzle driver capabilities */

static drv_caps_t drizzleclient_drv_caps =
{
  .multi_rows_insert = 1,
  .transactions = 1,
  .prepared_statements = 0,
  .auto_increment = 1,
  .serial = 0,
  .unsigned_int = 0,
  .all_need_fetch = 0,  
  .table_options_str = NULL
};



static drizzleclient_drv_args_t args;          /* driver args */

static char table_options_str[1024];   /* table options string */

static char use_ps; /* whether server-side prepared statemens should be used */

static sb_list_item_t *hosts_pos;

static pthread_mutex_t hosts_mutex;

/* Drizzle driver operations */

static int drizzleclient_drv_init(void);
static int drizzleclient_drv_describe(drv_caps_t *, const char *);
static int drizzleclient_drv_connect(db_conn_t *);
static int drizzleclient_drv_disconnect(db_conn_t *);
static int drizzleclient_drv_prepare(db_stmt_t *, const char *);
static int drizzleclient_drv_bind_param(db_stmt_t *, db_bind_t *, unsigned int);
static int drizzleclient_drv_bind_result(db_stmt_t *, db_bind_t *, unsigned int);
static int drizzleclient_drv_execute(db_stmt_t *, db_result_set_t *);
static int drizzleclient_drv_fetch(db_result_set_t *);
static int drizzleclient_drv_fetch_row(db_result_set_t *, db_row_t *);
static unsigned long long drizzleclient_drv_num_rows(db_result_set_t *);
static int drizzleclient_drv_query(db_conn_t *, const char *, db_result_set_t *);
static int drizzleclient_drv_free_results(db_result_set_t *);
static int drizzleclient_drv_close(db_stmt_t *);
static int drizzleclient_drv_store_results(db_result_set_t *);
static int drizzleclient_drv_done(void);

/* Drizzle driver definition */

static db_driver_t drizzleclient_driver =
{
  .sname = "drizzleclient",
  .lname = "Old Drizzle Client driver",
  .args = drizzleclient_drv_args,
  .ops =
  {
    drizzleclient_drv_init,
    drizzleclient_drv_describe,
    drizzleclient_drv_connect,
    drizzleclient_drv_disconnect,
    drizzleclient_drv_prepare,
    drizzleclient_drv_bind_param,
    drizzleclient_drv_bind_result,
    drizzleclient_drv_execute,
    drizzleclient_drv_fetch,
    drizzleclient_drv_fetch_row,
    drizzleclient_drv_num_rows,
    drizzleclient_drv_free_results,
    drizzleclient_drv_close,
    drizzleclient_drv_query,
    drizzleclient_drv_store_results,
    drizzleclient_drv_done
  },
  .listitem = {NULL, NULL}
};


/* Local functions */

static int parse_table_engine(const char *);
#ifdef HAVE_PS
static int get_drizzleclient_bind_type(db_bind_type_t);
#endif

/* Register Drizzle driver */


int register_driver_drizzleclient(sb_list_t *drivers)
{
  SB_LIST_ADD_TAIL(&drizzleclient_driver.listitem, drivers);

  return 0;
}


/* Drizzle driver initialization */


int drizzleclient_drv_init(void)
{
  char *s;
  
  args.hosts = sb_get_value_list("drizzle-host");
  if (SB_LIST_IS_EMPTY(args.hosts))
  {
    log_text(LOG_FATAL, "No Drizzle hosts specified, aborting");
    return 1;
  }
  hosts_pos = args.hosts;
  pthread_mutex_init(&hosts_mutex, NULL);
  
  args.port = (unsigned int)sb_get_value_int("drizzle-port");
  args.socket = sb_get_value_string("drizzle-socket");
  args.user = sb_get_value_string("drizzle-user");
  args.password = sb_get_value_string("drizzle-password");
  args.db = sb_get_value_string("drizzle-db");
  args.myisam_max_rows = (unsigned int)sb_get_value_int("myisam-max-rows");
  args.create_options = sb_get_value_string("drizzle-create-options");
  if (args.create_options == NULL)
    args.create_options = "";
  
  use_ps = 0;
#ifdef HAVE_PS
  drizzleclient_drv_caps.prepared_statements = 1;
  if (db_globals.ps_mode != DB_PS_MODE_DISABLE)
    use_ps = 1;
#endif

  s = sb_get_value_string("drizzle-engine-trx");
  if (s == NULL)
  {
    log_text(LOG_FATAL, "--drizzle-engine-trx requires an argument");
    return 1;
  }
  if (!strcasecmp(s, "yes"))
    args.engine_trx = ENGINE_TRX_YES;
  else if (!strcasecmp(s, "no"))
    args.engine_trx = ENGINE_TRX_NO;
  else if (!strcasecmp(s, "auto"))
    args.engine_trx = ENGINE_TRX_AUTO;
  else
  {
    log_text(LOG_FATAL, "Invalid value for drizzle-engine-trx: %s", s);
    return 1;
  }
  
  s = sb_get_value_string("drizzle-table-engine");

  return parse_table_engine(s);
}


/* Describe database capabilities (possibly depending on table type) */


int drizzleclient_drv_describe(drv_caps_t *caps, const char * table_name)
{
  db_conn_t    con;
  char         query[1024];
  DRIZZLE_RES    *res = NULL;
  DRIZZLE_ROW    row;
  DRIZZLE_FIELD  *fields;
  unsigned int num_fields;
  unsigned int i;
  int          connected = 0;
  int          rc = 0;
  
  *caps = drizzleclient_drv_caps;
  
  if (table_name == NULL)
    goto exit;

  /* Try to determine table type */
  if (drizzleclient_drv_connect(&con))
    goto error;

  /* Fix the hosts list */
  hosts_pos = args.hosts;
  
  connected = 1;
  snprintf(query, sizeof(query), "SHOW TABLE STATUS LIKE '%s'", table_name);
  
  rc = drizzleclient_real_query(con.ptr, query, strlen(query)); 
  DEBUG("drizzleclient_real_query(%p, \"%s\", %d) = %d", con.ptr, query, strlen(query), rc);
  if (rc)
    goto error;
  
  res = drizzleclient_store_result(con.ptr);
  DEBUG("drizzleclient_store_result(%p) = %p", con.ptr, res);
  if (res == NULL)
    goto error;

  num_fields = drizzleclient_num_fields(res);
  DEBUG("drizzleclient_num_fields(%p) = %d", res, num_fields);
  fields = drizzleclient_fetch_field(res);
  DEBUG("drizzleclient_fetch_field(%p) = %p", res, fields);
  for (i = 0; i < num_fields; i++)
    if (!strcasecmp(fields[i].name, "type") ||
        !strcasecmp(fields[i].name, "engine"))
      break;
  if (i >= num_fields)
    goto error;
  
  row = drizzleclient_fetch_row(res);
  DEBUG("drizzleclient_fetch_row(%p) = %p", res, row);
  if (row == NULL || row[i] == NULL)
    goto error;
  
  rc = parse_table_engine(row[i]);
  *caps = drizzleclient_drv_caps;
  goto exit;

 error:
  log_text(LOG_ALERT, "Error: failed to determine table '%s' type!", table_name);
  log_text(LOG_ALERT, "Drizzle error: %s", drizzleclient_error(con.ptr));
  rc = 1;
  
 exit:
  if (res != NULL)
  {
    drizzleclient_free_result(res);
    DEBUG("drizzleclient_free_result(%p)", res);
  }
  if (connected)
    drizzleclient_drv_disconnect(&con);

  return rc;
}


/* Connect to Drizzle database */


int drizzleclient_drv_connect(db_conn_t *sb_conn)
{
  DRIZZLE          *con;
  char           *host;
  
  con = (DRIZZLE *)malloc(sizeof(DRIZZLE));
  if (con == NULL)
    return 1;
  sb_conn->ptr = con;
  
  drizzleclient_create(con);
  DEBUG("drizzleclient_create(%p)", con);

  pthread_mutex_lock(&hosts_mutex);
  hosts_pos = SB_LIST_ITEM_NEXT(hosts_pos);
  if (hosts_pos == args.hosts)
    hosts_pos = SB_LIST_ITEM_NEXT(hosts_pos);
  host = SB_LIST_ENTRY(hosts_pos, value_t, listitem)->data;
  pthread_mutex_unlock(&hosts_mutex);
  
#if 0
  /*
    FIXME: the following leads to crash in the client lib.
    http://bugs.drizzle.com/?id=40552
  */
  drizzleclient_options(con, DRIZZLE_READ_DEFAULT_GROUP, "sysbench");
  DEBUG("drizzleclient_options(%p, DRIZZLE_READ_DEFAULT_GROUP, \"sysbench\")", con);
#endif
  
  
  DEBUG("drizzleclient_connect(%p, \"%s\", \"%s\", \"%s\", \"%s\", %u, \"%s\", %s)",
        con,
        host,
        args.user,
        args.password,
        args.db,
        args.port,
        args.socket,
        "CLIENT_MULTI_STATEMENTS"
        );
  if (!drizzleclient_connect(con,
                       host,
                       args.user,
                       args.password,
                       args.db,
                       args.port,
                       args.socket,
                       CLIENT_MULTI_STATEMENTS)
      )
  {
    log_text(LOG_FATAL, "unable to connect to Drizzle server, aborting...");
    log_text(LOG_FATAL, "error %d: %s", drizzleclient_errno(con),
           drizzleclient_error(con));
    free(con);
    return 1;
  }

  return 0;
}


/* Disconnect from Drizzle database */


int drizzleclient_drv_disconnect(db_conn_t *sb_conn)
{
  DRIZZLE *con = sb_conn->ptr;

  if (con != NULL)
  {
    DEBUG("drizzleclient_close(%p)", con);
    drizzleclient_close(con);
  }
  free(con);
  
  return 0;
}


/* Prepare statement */


int drizzleclient_drv_prepare(db_stmt_t *stmt, const char *query)
{
#ifdef HAVE_PS
  DRIZZLE      *con = (DRIZZLE *)stmt->connection->ptr;
  DRIZZLE_STMT *mystmt;
  unsigned int rc;

  if (con == NULL)
    return 1;

  if (use_ps)
  {
    mystmt = drizzleclient_stmt_init(con);
    DEBUG("drizzleclient_stmt_init(%p) = %p", con, mystmt);
    if (mystmt == NULL)
    {
      log_text(LOG_FATAL, "drizzleclient_stmt_init() failed");
      return 1;
    }
    stmt->ptr = (void *)mystmt;
    DEBUG("drizzleclient_stmt_prepare(%p, \"%s\", %d) = %p", mystmt, query, strlen(query), stmt->ptr);
    if (drizzleclient_stmt_prepare(mystmt, query, strlen(query)))
    {
      /* Check if this statement in not supported */
      rc = drizzleclient_errno(con);
      DEBUG("drizzleclient_errno(%p) = %u", con, rc);
      if (rc == ER_UNSUPPORTED_PS)
      {
        log_text(LOG_WARNING,
                 "Preparing of \"%s\" is unsupported, using emulation",
                 query);
        goto emulate;
      }
      else
      {
        log_text(LOG_FATAL, "drizzleclient_stmt_prepare() failed");
        log_text(LOG_FATAL, "Drizzle error: %d \"%s\"", rc,
                 drizzleclient_error(con));
        DEBUG("drizzleclient_stmt_close(%p)", mystmt);
        drizzleclient_stmt_close(mystmt);
        return 1;
      }
    }

    return 0;
  }

 emulate:
#endif /* HAVE_PS */

  /* Use client-side PS */
  stmt->emulated = 1;
  stmt->query = strdup(query);

  return 0;
}


/* Bind parameters for prepared statement */


int drizzleclient_drv_bind_param(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
{
  DRIZZLE        *con = (DRIZZLE *)stmt->connection->ptr;
#ifdef HAVE_PS
  DRIZZLE_BIND   *bind;
  unsigned int i;
  my_bool rc;
  unsigned long param_count;
#endif
  
  if (con == NULL)
    return 1;

#ifdef HAVE_PS
  if (!stmt->emulated)
  {
    if (stmt->ptr == NULL)
      return 1;
    /* Validate parameters count */
    param_count = drizzleclient_stmt_param_count(stmt->ptr);
    DEBUG("drizzleclient_stmt_param_count(%p) = %u", stmt->ptr, param_count);
    if (param_count != len)
      return 1;

    /* Convert SysBench bind structures to Drizzle ones */
    bind = (DRIZZLE_BIND *)calloc(len, sizeof(DRIZZLE_BIND));
    if (bind == NULL)
      return 1;
    for (i = 0; i < len; i++)
    {
      bind[i].buffer_type = get_drizzleclient_bind_type(params[i].type);
      bind[i].buffer = params[i].buffer;
      bind[i].buffer_length = params[i].max_len;
      bind[i].length = params[i].data_len;
      bind[i].is_null = params[i].is_null;
    }

    rc = drizzleclient_stmt_bind_param(stmt->ptr, bind);
    DEBUG("drizzleclient_stmt_bind_param(%p, %p) = %d", stmt->ptr, bind, rc);
    if (rc)
    {
      log_text(LOG_FATAL, "drizzleclient_stmt_bind_param() failed");
      log_text(LOG_FATAL, "Drizzle error: %d \"%s\"", drizzleclient_errno(con),
                 drizzleclient_error(con));
      free(bind);
      return 1;
    }
    free(bind);

    return 0;
  }
#endif /* HAVE_PS */

  /* Use emulation */
  if (stmt->bound_param != NULL)
    free(stmt->bound_param);
  stmt->bound_param = (db_bind_t *)malloc(len * sizeof(db_bind_t));
  if (stmt->bound_param == NULL)
    return 1;
  memcpy(stmt->bound_param, params, len * sizeof(db_bind_t));
  stmt->bound_param_len = len;
  
  return 0;

}


/* Bind results for prepared statement */


int drizzleclient_drv_bind_result(db_stmt_t *stmt, db_bind_t *params, unsigned int len)
{
#ifndef HAVE_PS
  /* No support for prepared statements */
  (void)stmt;   /* unused */
  (void)params; /* unused */
  (void)len;    /* unused */

  return 1;
#else
  DRIZZLE        *con = (DRIZZLE *)stmt->connection->ptr;
  DRIZZLE_BIND   *bind;
  unsigned int i;
  my_bool rc;
  
  if (con == NULL || stmt->ptr == NULL)
    return 1;

  /* Convert SysBench bind structures to Drizzle ones */
  bind = (DRIZZLE_BIND *)calloc(len, sizeof(DRIZZLE_BIND));
  if (bind == NULL)
    return 1;
  for (i = 0; i < len; i++)
  {
    bind[i].buffer_type = get_drizzleclient_bind_type(params[i].type);
    bind[i].buffer = params[i].buffer;
    bind[i].buffer_length = params[i].max_len;
    bind[i].length = params[i].data_len;
    bind[i].is_null = params[i].is_null;
  }

  rc = drizzleclient_stmt_bind_result(stmt->ptr, bind);
  DEBUG("drizzleclient_stmt_bind_result(%p, %p) = %d", stmt->ptr, bind, rc);
  if (rc)
  {
    free(bind);
    return 1;
  }
  free(bind);

  return 0;
#endif /* HAVE_PS */
}


/* Execute prepared statement */


int drizzleclient_drv_execute(db_stmt_t *stmt, db_result_set_t *rs)
{
  db_conn_t       *con = stmt->connection;
  char            *buf = NULL;
  unsigned int    buflen = 0;
  unsigned int    i, j, vcnt;
  char            need_realloc;
  int             n;
  unsigned int    rc;
 
#ifdef HAVE_PS
  (void)rs; /* unused */

  if (!stmt->emulated)
  {
    if (stmt->ptr == NULL)
    {
      log_text(LOG_DEBUG, "ERROR: exiting drizzleclient_drv_execute(), uninitialized statement");
      return SB_DB_ERROR_FAILED;
    }
    rc = (unsigned int)drizzleclient_stmt_execute(stmt->ptr);
    DEBUG("drizzleclient_stmt_execute(%p) = %u", stmt->ptr, rc);
    if (rc)
    {
      rc = drizzleclient_errno(con->ptr);
      DEBUG("drizzleclient_errno(%p) = %u", con->ptr, rc);
      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT || 
          rc == ER_CHECKREAD)
        return SB_DB_ERROR_DEADLOCK;
      log_text(LOG_ALERT, "failed to execute drizzleclient_stmt_execute(): Err%d %s",
               drizzleclient_errno(con->ptr),
               drizzleclient_error(con->ptr));
      return SB_DB_ERROR_FAILED;
    }
    return SB_DB_ERROR_NONE;
  }
#else
  (void)rc; /* unused */
#endif /* HAVE_PS */

  /* Use emulation */
  /* Build the actual query string from parameters list */
  need_realloc = 1;
  vcnt = 0;
  for (i = 0, j = 0; stmt->query[i] != '\0'; i++)
  {
  again:
    if (j+1 >= buflen || need_realloc)
    {
      buflen = (buflen > 0) ? buflen * 2 : 256;
      buf = realloc(buf, buflen);
      if (buf == NULL)
      {
        log_text(LOG_DEBUG, "ERROR: exiting drizzleclient_drv_execute(), memory allocation failure");
        return SB_DB_ERROR_FAILED;
      }
      need_realloc = 0;
    }

    if (stmt->query[i] != '?')
    {
      buf[j++] = stmt->query[i];
      continue;
    }

    n = db_print_value(stmt->bound_param + vcnt, buf + j, (int)(buflen - j));
    if (n < 0)
    {
      need_realloc = 1;
      goto again;
    }
    j += (unsigned int)n;
    vcnt++;
  }
  buf[j] = '\0';
  
  con->db_errno = drizzleclient_drv_query(con, buf, rs);
  free(buf);
  if (con->db_errno != SB_DB_ERROR_NONE)
  {
    log_text(LOG_DEBUG, "ERROR: exiting drizzleclient_drv_execute(), database error");
    return con->db_errno;
  }
  
  return SB_DB_ERROR_NONE;
}


/* Execute SQL query */


int drizzleclient_drv_query(db_conn_t *sb_conn, const char *query,
                      db_result_set_t *rs)
{
  DRIZZLE *con = sb_conn->ptr;
  unsigned int rc;

  (void)rs; /* unused */
  rc = (unsigned int)drizzleclient_real_query(con, query, strlen(query));
  DEBUG("drizzleclient_real_query(%p, \"%s\", %u) = %u", con, query, strlen(query), rc);
  if (rc)
  {
    rc = drizzleclient_errno(con);
    DEBUG("drizzleclient_errno(%p) = %u", con, rc);
    if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT || rc == ER_CHECKREAD)
      return SB_DB_ERROR_DEADLOCK;
    log_text(LOG_ALERT, "failed to execute Drizzle query: `%s`:", query);
    log_text(LOG_ALERT, "Error %d %s", drizzleclient_errno(con), drizzleclient_error(con));
    return SB_DB_ERROR_FAILED; 
  }  

  return SB_DB_ERROR_NONE;
}


/* Fetch row from result set of a prepared statement */


int drizzleclient_drv_fetch(db_result_set_t *rs)
{
  /* NYI */
  (void)rs;

  return 1;
}


/* Fetch row from result set of a query */


int drizzleclient_drv_fetch_row(db_result_set_t *rs, db_row_t *row)
{
  /* NYI */
  (void)rs;  /* unused */
  (void)row; /* unused */
  
  return 1;
}


/* Return the number of rows in a result set */


unsigned long long drizzleclient_drv_num_rows(db_result_set_t *rs)
{
  return rs->nrows;
}


/* Store results from the last query */


int drizzleclient_drv_store_results(db_result_set_t *rs)
{
  DRIZZLE        *con = rs->connection->ptr;
  DRIZZLE_RES    *res;
  DRIZZLE_ROW    row;
  unsigned int rc;

#ifdef HAVE_PS
  /* Is this result set from prepared statement? */
  if (rs->statement != NULL && rs->statement->emulated == 0)
  {
    if (rs->statement->ptr == NULL)
      return 1;

    rc = (unsigned int)drizzleclient_stmt_store_result(rs->statement->ptr);
    DEBUG("drizzleclient_stmt_store_result(%p) = %d", rs->statement->ptr, rc);
    if (rc)
    {
      rc = drizzleclient_errno(con);
      DEBUG("drizzleclient_errno(%p) = %d", con, rc);
      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT)
      {
        log_text(LOG_WARNING,
                 "drizzleclient_stmt_store_result() failed with error: (%d) %s", rc,
                 drizzleclient_error(con));
        return SB_DB_ERROR_DEADLOCK;
      }

      log_text(LOG_ALERT, "Drizzle error: %s\n", drizzleclient_error(con));
      return SB_DB_ERROR_FAILED;
    }
    rs->nrows = drizzleclient_stmt_num_rows(rs->statement->ptr);
    DEBUG("drizzleclient_stmt_num_rows(%p) = %d", rs->statement->ptr, rs->nrows);
    do {
      rc = (unsigned int)drizzleclient_stmt_fetch(rs->statement->ptr);
      DEBUG("drizzleclient_stmt_fetch(%p) = %d", rs->statement->ptr, rc);
    } while(rc == 0);

    return SB_DB_ERROR_NONE;
  }
#endif
  
  if (con == NULL)
    return SB_DB_ERROR_FAILED;
  
  /* using store results for speed will not work for large sets */
  res = drizzleclient_store_result(con);
  DEBUG("drizzleclient_store_result(%p) = %p", con, res);
  if (res == NULL)
  {
      rc = drizzleclient_errno(con);
      DEBUG("drizzleclient_errno(%p) = %u", con, rc);
      if (rc == ER_LOCK_DEADLOCK || rc == ER_LOCK_WAIT_TIMEOUT)
      {
        log_text(LOG_WARNING,
                 "drizzleclient_store_result() failed with error: (%u) %s", rc,
                 drizzleclient_error(con));
        return SB_DB_ERROR_DEADLOCK;
      }
    log_text(LOG_ALERT, "Drizzle error: %s", drizzleclient_error(con));
    return SB_DB_ERROR_FAILED; 
  }
  rs->ptr = (void *)res;

  rs->nrows = drizzleclient_num_rows(res);
  DEBUG("drizzleclient_num_rows(%p) = %u", res, rs->nrows);
  
  /* just fetch result */
  while((row = drizzleclient_fetch_row(res)))
    DEBUG("drizzleclient_fetch_row(%p) = %p", res, row);

  return SB_DB_ERROR_NONE;
}


/* Free result set */


int drizzleclient_drv_free_results(db_result_set_t *rs)
{
#ifdef HAVE_PS
  /* Is this a result set of a prepared statement */
  if (rs->statement != NULL && rs->statement->emulated == 0)
  {
    DEBUG("drizzleclient_stmt_free_result(%p)", rs->statement->ptr);
    return drizzleclient_stmt_free_result(rs->statement->ptr);
  }
#endif

  if (rs->ptr != NULL)
  {
    DEBUG("drizzleclient_free_result(%p)", rs->ptr);
    drizzleclient_free_result((DRIZZLE_RES *)rs->ptr);
    return 0;
  }

  return 1;
}


/* Close prepared statement */


int drizzleclient_drv_close(db_stmt_t *stmt)
{
#ifdef HAVE_PS
  if (stmt->ptr == NULL)
    return 1;

  DEBUG("drizzleclient_stmt_close(%p)", stmt->ptr);
  return drizzleclient_stmt_close(stmt->ptr);
#else
  (void)stmt; /* unused */
  return 0;
#endif
}


/* Uninitialize driver */
int drizzleclient_drv_done(void)
{
  return 0;
}


/* Parse table type and set driver capabilities */


int parse_table_engine(const char *type)
{
  /* Determine if the engine supports transactions */
  if (args.engine_trx == ENGINE_TRX_AUTO)
  {
    if (!strcasecmp(type, "myisam") || !strcasecmp(type, "heap") ||
        !strcasecmp(type, "maria"))
      drizzleclient_drv_caps.transactions = 0;
    else if (!strcasecmp(type, "innodb") ||
             !strcasecmp(type, "bdb") || !strcasecmp(type, "berkeleydb") ||
             !strcasecmp(type, "ndbcluster") ||
             !strcasecmp(type, "federated") ||
             !strcasecmp(type, "pbxt") ||
             !strcasecmp(type, "falcon"))
      drizzleclient_drv_caps.transactions = 1;
    else
    {
      log_text(LOG_FATAL, "Failed to determine transactions support for "
               "unknown storage engine '%s'", type);
      log_text(LOG_FATAL, "Specify explicitly with --drizzle-engine-trx option");
      return 1;
    }
  }
  else
    drizzleclient_drv_caps.transactions = (args.engine_trx == ENGINE_TRX_YES);

  /* Get engine-specific options string */
  if (!strcasecmp(type, "myisam"))
  {
    if (args.myisam_max_rows > 0)
      snprintf(table_options_str, sizeof(table_options_str),
               "%s  "ENGINE_CLAUSE"=MyISAM MAX_ROWS=%u ",
               args.create_options,
               args.myisam_max_rows);
    else
      snprintf(table_options_str, sizeof(table_options_str),
               "%s "ENGINE_CLAUSE"=MyISAM", args.create_options);
  }
  else if (!strcasecmp(type, "ndbcluster"))
    snprintf(table_options_str, sizeof(table_options_str),
             "%s  "ENGINE_CLAUSE"=NDB", args.create_options);
  else
    snprintf(table_options_str, sizeof(table_options_str),
             "%s  "ENGINE_CLAUSE"=%s", args.create_options, type);

  drizzleclient_drv_caps.table_options_str = table_options_str;
  
  return 0;
}


#ifdef HAVE_PS
/* Map SQL data type to bind_type value in DRIZZLE_BIND */


int get_drizzleclient_bind_type(db_bind_type_t type)
{
  unsigned int i;

  for (i = 0; db_drizzleclient_bind_map[i].db_type != DB_TYPE_NONE; i++)
    if (db_drizzleclient_bind_map[i].db_type == type)
      return db_drizzleclient_bind_map[i].my_type;

  return -1;
}

#endif /* HAVE_PS */
